/*********************************************************************NVMH3****

Copyright NVIDIA Corporation 2002
TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
*AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.


Comments:

******************************************************************************/

//
// demo.cpp
//
// basic example of the use of the Cg runtime in a simple OpenGL program
//

#include <iostream>
#include <fstream>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <OpenGL/glext.h>
#include <GLUT/glut.h>
#else
#include "extgl.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#include <GL/glut.h>
#endif

#ifndef CWD
#define CWD ""
#endif

#define RES 512

// globals
int		lod=4;
double		dx=0.0,dy=0.0;

GLuint		vphandle;
GLuint		fphandle;

// global texture ids
const int	textures=2;
unsigned long	texID[] = {GL_TEXTURE0_ARB,GL_TEXTURE1_ARB};
GLuint		texNames[2];

// forward declarations of functions
int	main(int argc, char *argv[]);
void	interpmenuhandler(int c);
void	lodmenuhandler(int l);
void	mainmenuhandler(int m);
void	mousemove(int x, int y);
void	keyboard(unsigned char key, int x, int y);
void	display();
void	initializeGL();
void	loadARBPrograms();
void	imageTexture();
void	maskTexture();
float	gauss(int i, int j, float sigma);
float	width();
float	height();

int main(int argc, char *argv[])
{
	int	mainmenu,interpmenu,lodmenu;
	
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutInitWindowSize(RES, RES);
  glutCreateWindow("ARB gcd");

  interpmenu = glutCreateMenu(interpmenuhandler);
  glutAddMenuEntry("None",1);
  glutAddMenuEntry("Intra-map",2);
  glutAddMenuEntry("Inter-map",3);
  glutAddMenuEntry("Both",4);

  lodmenu = glutCreateMenu(lodmenuhandler);
  glutAddMenuEntry("1",1);
  glutAddMenuEntry("2",2);
  glutAddMenuEntry("3",3);
  glutAddMenuEntry("4",4);

  mainmenu = glutCreateMenu(mainmenuhandler);
  glutAddSubMenu("Interpolation",interpmenu);
  glutAddSubMenu("LOD",lodmenu);
  glutAddMenuEntry("Quit",9);
// GLUT_RIGHT_BUTTON doesn't work on Macs unless mouse emulation is on
//glutAttachMenu(GLUT_RIGHT_BUTTON);
  glutAttachMenu(GLUT_LEFT_BUTTON);

//glutMotionFunc(mousemove);
  glutPassiveMotionFunc(mousemove);
  glutKeyboardFunc(keyboard);
  glutDisplayFunc(display);

  initializeGL();

  // and all the rest happens in the display function...
  glutMainLoop();

  return 0;
}

void mainmenuhandler(int m)
{
  switch(m) {
    case 9:
      glDeleteProgramsARB(1,&vphandle);
      glDeleteProgramsARB(1,&fphandle);
      exit(0);
    break;
  }
}

void interpmenuhandler(int c)
{
//fprintf(stderr,"interp: %d\n",c);
  //                                                       (Manhattan dist.)
  //                                   intra-map interp    inter-map interp
  // GL_TEXTURE_MIN_FILTER options:
  //   GL_NEAREST_MIPMAP_NEAREST:             no                  no
  //   GL_LINEAR_MIPMAP_NEAREST              yes                  no
  //   GL_NEAREST_MIPMAP_LINEAR               no                 yes
  //   GL_LINEAR_MIPMAP_LINEAR               yes                 yes
  // GL_TEXTURE_MAG_FILTER options:
  //   GL_NEAREST                             no                 n/a
  //   GL_LINEAR                             yes                 n/a
  glActiveTextureARB(texID[0]);
  switch(c) {
    case 1:
      // no smoothing
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_NEAREST_MIPMAP_NEAREST);
      break;
    case 2:
      // intra-map
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_LINEAR_MIPMAP_NEAREST);
      break;
    case 3:
      // inter-map
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_NEAREST_MIPMAP_LINEAR);
      break;
    case 4:
    default:
      // both
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_LINEAR_MIPMAP_LINEAR);
      break;
  }

  glutPostRedisplay();
}

void lodmenuhandler(int l)
{
  lod = l;
  glutPostRedisplay();
}

void mousemove(int x, int y)
{ 
//fprintf(stderr,"mousemove: (%d,%d)\n",x,y);
  dx = width()/2.0 - (float)x;
  dy = (float)y - height()/2.0;
  dx /= width();
  dy /= height();
//fprintf(stderr,"mousemove: (%4.2f,%4.2f)\n",(float)dx,(float)dy);
  glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
  switch(key) {
    case 'q':
    case 'Q':
    case 27:
      glDeleteProgramsARB(1,&vphandle);
      glDeleteProgramsARB(1,&fphandle);
      exit(0);
      break;
    case '1':
      lod = 1;
      break;
    case '2':
      lod = 2;
      break;
    case '3':
      lod = 3;
      break;
    case '4':
      lod = 4;
      break;
  }
  glutPostRedisplay();
}

void initializeGL()
{
  fprintf(stderr,"GL: %s\n",glGetString(GL_VERSION));

  // check extensions
#ifndef __APPLE__
  extgl_Initialize();

  // make sure extensions we need are present
  if( !(extgl_ExtensionSupported( "GL_ARB_vertex_program" )) ) {
      fprintf(stderr,"error: vertex program support is required\n");
      exit(1);
  }
  if( !(extgl_ExtensionSupported( "GL_ARB_fragment_program" )) ) {
      fprintf(stderr,"error: fragment program support is required\n");
      exit(1);
  }
#endif

  // should check for GL_ARB_fragment_program extension support
  loadARBPrograms();

  glGenTextures(textures,texNames);
  imageTexture();
  maskTexture();

  glClearColor(.25, .25, .25, 1.);

  glDisable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);
}

void display() 
{
	float	lx,ly;
	float	lw,lh;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0,width(),0.0,height());

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glDrawBuffer(GL_BACK);
  glClear(GL_COLOR_BUFFER_BIT);

  glActiveTextureARB(texID[0]);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, texNames[0]);

  glActiveTextureARB(texID[1]);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, texNames[1]);

  glEnable(GL_VERTEX_PROGRAM_ARB);
  glBindProgramARB(GL_VERTEX_PROGRAM_ARB,vphandle);
  glEnable(GL_FRAGMENT_PROGRAM_ARB);
  glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB,fphandle);

  // set fragment program local parameter
//glProgramLocalParameter1fARB(GL_FRAGMENT_PROGRAM_ARB,0,lod);
  glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB,0,lod,0.0,0.0,0.0);

  lx = (float)dx;
  ly = (float)dy;
//fprintf(stderr,"(lx,ly): (%4.2f,%4.2f)\n",lx,ly);
  // calculate mask 'aspect ratio': if mask is made to match window
  // size, (1,1) gives the proper offset
//lw = (float)width() / mask->cols();
//lh = (float)height() / mask->rows();
  lw = 1.0;
  lh = 1.0;
//fprintf(stderr,"(lw,lh): (%4.0f,%4.0f)\n",lw,lh);

  // render
  glBegin(GL_QUADS);
  {
    glMultiTexCoord2fARB(texID[0],0,0);
    glMultiTexCoord2fARB(texID[1],lx,ly);
    glVertex2f(0,0);

    glMultiTexCoord2fARB(texID[0],1,0);
    glMultiTexCoord2fARB(texID[1],lx+lw,ly);
    glVertex2f(width(),0);

    glMultiTexCoord2fARB(texID[0],1,1);
    glMultiTexCoord2fARB(texID[1],lx+lw,ly+lh);
    glVertex2f(width(),height());

    glMultiTexCoord2fARB(texID[0],0,1);
    glMultiTexCoord2fARB(texID[1],lx,ly+lh);
    glVertex2f(0,height());
  }
  glEnd();

  glDisable(GL_FRAGMENT_PROGRAM_ARB);
  glDisable(GL_VERTEX_PROGRAM_ARB);

  glActiveTextureARB(texID[1]);
  glDisable(GL_TEXTURE_2D);

  glActiveTextureARB(texID[0]);
  glDisable(GL_TEXTURE_2D);

  glutSwapBuffers();
  glFlush();
}

void loadARBPrograms()
{
	std::ifstream file;
	std::ofstream ofs;

  // load and compile fragment program
  file.open( CWD "vert.vp" );
  std::istreambuf_iterator<char> vbegin( file ), vend;
  std::string vdata( vbegin, vend );
  file.close();

  // debugging
//ofs.open( CWD "vert-copy.vp" );
//ofs << vdata.c_str();
//ofs.close();

  // generate and bind vp handle
  glGenProgramsARB(1,&vphandle);
  glBindProgramARB(GL_VERTEX_PROGRAM_ARB,vphandle);

  // upload the program
  glProgramStringARB(GL_VERTEX_PROGRAM_ARB,
                     GL_PROGRAM_FORMAT_ASCII_ARB,
                     vdata.size(), vdata.c_str());

  // check for compile errors
  if(glGetError() == GL_INVALID_OPERATION) {
    // find the error position
    GLint pos;
    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB,&pos);
    // print error message
    const GLubyte *error = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
    std::cerr << "error: compile error: " << pos << ": " << error << std::endl;
    exit(1);
  }
  glDisable(GL_VERTEX_PROGRAM_ARB);

  // load and compile fragment program
  file.open( CWD "frag.fp" );
  std::istreambuf_iterator<char> fbegin( file ), fend;
  std::string fdata( fbegin, fend );

  // generate and bind fp handle
  glGenProgramsARB(1,&fphandle);
  glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB,fphandle);

  // upload the program
  glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB,
                     GL_PROGRAM_FORMAT_ASCII_ARB,
                     fdata.size(), fdata.c_str());

  // check for compile errors
  if(glGetError() == GL_INVALID_OPERATION) {
    // find the error position
    GLint pos;
    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB,&pos);
    // print error message
    const GLubyte *error = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
    std::cerr << "error: compile error: " << pos << ": " << error << std::endl;
    exit(1);
  }
  glDisable(GL_FRAGMENT_PROGRAM_ARB);
}

void imageTexture()
{
	GLint	error;
	float	*image = new float[RES*RES*4];
	float	*ip = image;

  srand((unsigned int)0x1337);
  for(int i = 0; i < RES; ++i) {
    for(int j = 0; j < RES; ++j) {
      *ip++ = (float)((float)rand() / (float)RAND_MAX);
      *ip++ = (float)((float)rand() / (float)RAND_MAX);
      *ip++ = (float)((float)rand() / (float)RAND_MAX);
      *ip++ = (float)((float)rand() / (float)RAND_MAX);
    }
  }
  glBindTexture(GL_TEXTURE_2D, texNames[0]);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  error = gluBuild2DMipmaps(GL_TEXTURE_2D,
                            GL_RGBA,		// internalFormat
                            RES,RES,		// width,height
                            GL_RGBA,		// format
                            GL_FLOAT,		// type
                            image);		// texels
  if(error)	fprintf(stderr,"%s\n",gluErrorString(error));
  else		fprintf(stderr,"gluBuild2DMipmaps succeeded.\n");

  delete [] image;
}

void maskTexture()
{
	float	g,sigma=100.0;
	float	*mask = new float[RES*RES*4];
	float	*ip = mask;

  for(int i = 0; i < RES; ++i) {
    for(int j = 0; j < RES; ++j) {
      g = gauss(i-RES/2,j-RES/2,sigma);
      *ip++ = g;
      *ip++ = g;
      *ip++ = g;
      *ip++ = g;
    }
  }
  glBindTexture(GL_TEXTURE_2D, texNames[1]);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,RES,RES,0,GL_RGBA,GL_FLOAT,mask);

  delete [] mask;
}

float gauss(int i, int j, float sigma)
{
  return(exp((i*i + j*j)/(-2.0*sigma*sigma)));
}

float width()
{
  return(glutGet(GLUT_WINDOW_WIDTH));
}

float height()
{
  return(glutGet(GLUT_WINDOW_HEIGHT));
}
